/*
Bibliothèque de fonctions pour calculer des Transformées de Fourier

Copyright (C) 2004 Observatoire de Paris-Meudon

Ce programme est un logiciel libre ; vous pouvez le redistribuer et/ou le modifier conformément aux dispositions de la Licence Publique Générale GNU, telle que publiée par la Free Software Foundation ; version 2 de la licence, ou encore (à votre choix) toute version ultérieure.

Ce programme est distribué dans l'espoir qu'il sera utile, mais SANS AUCUNE GARANTIE ; sans même la garantie implicite de COMMERCIALISATION ou D'ADAPTATION A UN OBJET PARTICULIER. Pour plus de détail, voir la Licence Publique Générale GNU.

Vous devez avoir reçu un exemplaire de la Licence Publique Générale GNU en même temps que ce programme ; si ce n'est pas le cas, écrivez à la Free Software Foundation Inc., 675 Mass Ave, Cambridge, MA 02139, Etats-Unis.
*/

package tflib;

/**
 * Bibliothèque de fonctions pour calculer des Transformées de Fourier
 * encodage: UTF-8
 */
public class TFLib {
    
	protected double[][] tw_cos;
	protected double[][] tw_sin;
	protected double[] f_r;
	protected double[] f_i;
	
	protected TFLib(double[] f_r, double[] f_i) {
		this.f_r = f_r;
		this.f_i = f_i;
	}
	
	/**
	 * FFT en 2D
     * @param imagep    image en entrée
     * @param figured   image en sortie
     * @param log       log sur l'image de sortie
     * @param quadrants inverser les quadrants
	 */
    public static void transfo2D_FFT(double[][] imagep, double[][] figured, boolean log, boolean quadrants) {
		int dimx = imagep[0].length;
		int dimy = imagep.length;
		for (int i=0; i<dimx; i++)
			for (int j=0; j<dimy; j++)
				figured[i][j] = 0;
		double[][][] fk = new double[dimx][dimy][2];
		for (int kx=0; kx<dimx; kx++) {
			double[] f_r = new double[dimy];
			double[] f_i = new double[dimy];
			for (int i=0; i<dimy; i++) {
				f_r[i] = imagep[kx][i];
				f_i[i] = 0;
			}
			
			fft_rec(false, f_r, f_i);
			
			for (int i=0; i<dimy; i++) {
				fk[kx][i][0] = f_r[i];
				fk[kx][i][1] = f_i[i];
			}
		}
		for (int ky=0; ky<dimy; ky++) {
			double[] f_r = new double[dimx];
			double[] f_i = new double[dimx];
			for (int i=0; i<dimx; i++) {
				f_r[i] = fk[i][ky][0];
				f_i[i] = fk[i][ky][1];
			}
			
			fft_rec(false, f_r, f_i);
			
			double v;
			for (int i=0; i<dimx; i++) {
				if (log)
					v = Math.log(Math.sqrt(f_r[i]*f_r[i] + f_i[i]*f_i[i]));
				else
					v = Math.sqrt(f_r[i]*f_r[i] + f_i[i]*f_i[i]);
				if (quadrants) {
					// réorganisation des coordonnées
					int i2;
					if (i >= dimx/2)
						i2 = i - dimx/2;
					else
						i2 = i + dimx/2;
					int ky2;
					if (ky >= dimy/2)
						ky2 = ky - dimy/2;
					else
						ky2 = ky + dimy/2;
					figured[i2][ky2] = v;
				} else
					figured[i][ky] = v;
			}
		}
	}
	
	protected void bit_reverse() {
		bit_reverse(f_r);
		bit_reverse(f_i);
	}
	
	protected static void bit_reverse(double[] t) {
		int j = 0;
		double temp;
		int m;
		for (int i=0; i<t.length-1; i++) {
			if (j > i) {
				temp = t[j];
				t[j] = t[i];
				t[i] = temp;
			}
			m = t.length / 2;
			while (j >= m) {
				j -= m;
				m /= 2;
			}
			j += m;
		}
	}
	
	/**
	 * FFT récursive un peu optimisée.
	 * la taille de f (f_r et f_i) doit être une puissance de 2.
	 * avec precalcul_tw.
	 */
	public static void fft_rec(boolean inverse, double[] f_r, double[] f_i) {
		TFLib lib = new TFLib(f_r, f_i);
		lib.precalcul_tw();
		int log2n = (int)(Math.log(f_r.length)/Math.log(2));
		lib.dif(inverse, 0, log2n);
		lib.bit_reverse();
	}
	
	protected void precalcul_tw() {
		int n = f_r.length;
		int log2n = (int)(Math.log(n)/Math.log(2));
		tw_cos = new double[n/2][log2n+1];
		tw_sin = new double[n/2][log2n+1];
		double tw;
		int m = n;
		int log2m = log2n;
		while (m > 1) {
			for (int i=0; i<m/2; i++) {
				tw = 2*Math.PI*((double)i)/m;
				tw_cos[i][log2m] = Math.cos(tw);
				tw_sin[i][log2m] = Math.sin(tw);
			}
			m /= 2;
			log2m--;
		}
	}
	
	protected void dif(boolean inverse, int ind, int log2n) {
		int np = 1<< (log2n-1);
		int ind2 = ind + np;
		double e_r, e_i, o_r, o_i;
		double tw_r, tw_i, tmp_r, tmp_i;
		
		for (int i=0; i<np; i++) {
			e_r = f_r[ind+i] + f_r[ind2+i];
			e_i = f_i[ind+i] + f_i[ind2+i];
			
			tw_r = tw_cos[i][log2n];
			tw_i = -tw_sin[i][log2n];
			if (inverse)
				tw_i = -tw_i;
			
			tmp_r = f_r[ind+i] - f_r[ind2+i];
			tmp_i = f_i[ind+i] - f_i[ind2+i];
			
			o_r = tmp_r * tw_r - tmp_i * tw_i;
			o_i = tmp_r * tw_i + tmp_i * tw_r;
			
			f_r[ind+i] = e_r;
			f_i[ind+i] = e_i;
			
			f_r[ind2+i] = o_r;
			f_i[ind2+i] = o_i;
		}
		
		if (log2n > 1) {
			dif(inverse, ind, log2n-1);
			dif(inverse, ind2, log2n-1);
		}
	}
	
}